#include "activelz.hpp"
#include "log.hpp"

#include <iostream>
#include <fstream>

// boost includes
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>


// vacuous action assumed to generate first observation
const action_t NullAction = -1;

// uniform [0..1) random number generation
randsrc_t rng;
randgen_t urng(rng);


/* create a new LZ context node */
LZContextNode::LZContextNode(
	LZContextNode *p, 
	size_t n_obs, 
	size_t n_actions, 
	const observation_t &o, 
	action_t a
) :
	m_parent(p), 
	m_prob(1.0 / double(n_obs)), 
	m_count(0), 
	m_cost_to_go(0.0),
	m_obs(o),
	m_action(a)
{
}


/* add a new (assumed unique) child node */
LZContextNode *LZContextNode::addChild(LZContextNode *n) {

	assert(n != NULL);
	assert(getChild(n->m_obs, n->m_action) == NULL);

	m_children.push_back(n);
	
	return n;
}


/* how many times has this context been seen before? */
uint64_t LZContextNode::count(void) const {

	return m_count;
}


/* computes Pr(obs_n | a_{1:n-1} obs_{1:n-1} */
double LZContextNode::prob(void) const {

	assert(m_prob >= 0.0 && m_prob <= 1.0);

	return m_prob;
}


/* do we have no children */
bool LZContextNode::isLeaf(void) const {

	return m_children.empty();
}


/* find the leaf node corresponding to <obs_n, action_{n-1}> */
LZContextNode *LZContextNode::getChild(const observation_t &o, action_t a) {

	for (size_t i=0; i < m_children.size(); ++i) {
		if (m_children[i].m_obs == o && m_children[i].m_action == a) {
			return &m_children[i];
		}
	}
	return NULL;
}



/* Lempel-Ziv Context Tree */
LZContextTree::LZContextTree(
	unsigned int obs_bits,
	unsigned int rew_bits,
	unsigned int n_actions,
	double discount) :
	m_root(NULL, 1 << (obs_bits + rew_bits), n_actions, observation_t(), NullAction),
	m_obs_bits(obs_bits),
	m_rew_bits(rew_bits),
	m_n_actions(n_actions),
	m_discount(discount)
{
}


/* create LZ-Context Tree from a file */
LZContextTree::LZContextTree(const std::string &filename, bool use_binary_io) {

	if (use_binary_io) {

		// load the agent from file
		std::ifstream ifs(filename.c_str(), std::ios_base::binary);
		if (!ifs) throw std::invalid_argument("bad Active-LZ agent file.");
		LINFO_ << "loading (binary-mode) Active-lZ agent from: " << filename << std::endl;

		boost::archive::binary_iarchive ia(ifs);
        ia >> *this;

	} else { 
		
		// load the agent from file
		std::ifstream ifs(filename.c_str());
		if (!ifs) throw std::invalid_argument("bad Active-LZ agent file.");
		LINFO_ << "loading (text-mode) Active-LZ agent from: " << filename << std::endl;

		boost::archive::text_iarchive ia(ifs);
        ia >> *this;
    }
}


/* how many times has this context been seen before? */
uint64_t LZContextTree::count(const context_t &context) {

	LZContextNode *n = findNode(context);
	return n != NULL ? n->count() : 0;
}


/* determine the estimated best action, breaking ties at random */
action_t LZContextTree::bestAction(const context_t &context) {

	LZContextNode *n = findNode(context);

	double best = std::numeric_limits<double>::infinity();
	std::vector<action_t> best_a;

	for (action_t a=0; a < m_n_actions; a++) {
		double c = cost(n, a);
		if (c < best) best_a.clear(); 
		if (c <= best) {
			best = c;
			best_a.push_back(a); 
		}
	}

	assert(best_a.size() > 0);
	size_t idx = size_t(urng() * double(best_a.size()));
	return best_a[idx];
}


/* attempts to match a context within the tree, NULL on no match */
LZContextNode *LZContextTree::findNode(const context_t &context) {

	LZContextNode *r = &m_root;

	context_t::const_iterator i = context.begin();
	for (; i != context.end(); ++i) {
		r = r->getChild(i->obs, i->action);
		if (r == NULL) return NULL;
	}
	return r;
}


/* update the tree with new context information */
void LZContextTree::update(const context_t &context) {

	assert(context.size() > 0);
	assert(findNode(context) == NULL); // context must not exist

	LZContextNode *n = &m_root;
	
	// match the existing subcontext
	context_t::const_iterator i = context.begin();
	for (; i != context.end(); ++i) {
		LZContextNode *next = n->getChild(i->obs, i->action);
		if (next == NULL) break;
		n = next;
	}

	// extend the context out for the new part
	for (; i != context.end(); ++i) {
		n = n->addChild(new LZContextNode(n, nObservations(), 
				m_n_actions, i->obs, i->action));
	}

	double num_bias   = 0.5;
	double denom_bias = double(nObservations()) * 0.5;

	for (; n != NULL; n = n->m_parent) {
	
		// update context count
		n->m_count++;
		
		// update probability estimates	
		double denom = 0.0;
		LZContextNode *p = n->m_parent;

		if (p != NULL) {
			for (size_t k=0; k < p->m_children.size(); k++) {
				const LZContextNode &c = p->m_children[k];	
				if (c.m_action == n->m_action) denom += double(c.m_count);
			}
			for (size_t k=0; k < p->m_children.size(); k++) {
				LZContextNode &c = p->m_children[k];	
				if (c.m_action == n->m_action) {
					c.m_prob = (double(c.m_count) + num_bias) / (denom + denom_bias);
					assert(c.m_prob >= 0.0 && c.m_prob <= 1.0);
				}
			}
		} 

		// update cost to go estimates
		double best = std::numeric_limits<double>::infinity();
		for (action_t a = 0; a < m_n_actions; a++) {
			double g = cost(n, a);
			if (g < best) best = g;
		}
		n->m_cost_to_go = best;
	}
}


/* expected future cost if action a is executed from a certain context */
double LZContextTree::cost(LZContextNode *n, action_t a) const {

	if (n == NULL) return 0.0;

	double denom = 0.0;
	for (size_t i=0; i < n->m_children.size(); i++) {
		const LZContextNode &ch = n->m_children[i];
		if (ch.m_action == a) denom += double(ch.m_count);
	}
	denom += double(nObservations()) * 0.5;

	double total = 0.0;
	
	for (size_t o=0; o < nObservations(); o++) {
		
		observation_t obs(m_obs_bits + m_rew_bits, o);
		obs.resize(m_obs_bits + m_rew_bits);

		LZContextNode *c = n->getChild(obs, a);
		double g = costFromPercept(obs);

		if (c != NULL) {
			total += c->m_prob * (g + m_discount * c->m_cost_to_go);
		} else {
			total += (1.0 / denom) * g;
		}
	}
	return total;
}


/* extract the cost component from a percept */
double LZContextTree::costFromPercept(const observation_t &obs) const {

	assert(obs.size() == m_obs_bits + m_rew_bits);

	// split into observation and reward components
	boost::dynamic_bitset<> rew = (obs << m_obs_bits) >> m_obs_bits;
	rew.resize(m_rew_bits); 
	double reward = double(rew.to_ulong());

	return -reward; // cost is negative reward
}


/* the estimated future discounted cost from a context */
double LZContextTree::costToGo(const context_t &context, action_t a) {

	LZContextNode *n = findNode(context);
	return cost(n, a);
}


/* number of distinct observations */
size_t LZContextTree::nObservations(void) const {

	size_t rval = 1;
	return rval << (m_obs_bits + m_rew_bits);
}


/* computes Pr(obs_n | a_{1:n-1} obs_{1:n-1} */
double LZContextTree::probFromCounts(const LZContextNode &n) const  {

	double num   = double(n.count()) + 0.5;
	double denom = double(nObservations()) * 0.5;

	LZContextNode *p = n.m_parent;
	if (p == NULL) return num / (denom + double(n.count()));

	for (size_t i=0; i < p->m_children.size(); i++) {
		const LZContextNode &c = p->m_children[i];
		if (c.m_action == n.m_action) denom += c.count();
	}
	
	return num / denom;
}
