#ifndef __ACTIVELZ_HPP__
#define __ACTIVELZ_HPP__

#include <exception>
#include <vector>

// boost includes
#include <boost/cstdint.hpp>
#include <boost/random.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <boost/serialization/dynamic_bitset.hpp> // unofficial

// fixed size integer types
using boost::int16_t;
using boost::uint16_t;
using boost::uint32_t;
using boost::int32_t;
using boost::uint64_t;
using boost::int64_t;


// program configuration options
namespace boost {
	namespace program_options {
		class variables_map;
	}
}
extern boost::program_options::variables_map options;

class LZContextNode;
class LZContextTree;

// describes an agent action
typedef unsigned int action_t;

// observation type
typedef boost::dynamic_bitset<> observation_t;

// context type
struct obs_action_t {
	obs_action_t(const action_t &a, const observation_t &o) 
		: action(a), obs(o) {}
	action_t action;
	observation_t obs;
};
typedef std::vector<obs_action_t> context_t;

// describe the reward accumulated by an agent
typedef double reward_t;

// describe the age of an agent
typedef unsigned int age_t;

// random number generator to supply noise
typedef boost::mt19937 randsrc_t;
typedef boost::uniform_01<randsrc_t> randgen_t;

// a list of context nodes
typedef boost::ptr_vector<LZContextNode> lznode_list_t;

// vacuous action assumed to generate first observation
extern const action_t NullAction;

// uniform [0..1) random number generation
extern randsrc_t rng;
extern randgen_t urng;


// Context Node
class LZContextNode {

	friend LZContextTree;
	friend class boost::serialization::access;

	public:

		LZContextNode(LZContextNode *parent, size_t n_obs, 
			size_t n_actions, const observation_t &obs, action_t a);

		// how many times has this context been seen before?
		uint64_t count(void) const;

		// find the leaf node corresponding to <obs_n, action_{n-1}>
		LZContextNode *getChild(const observation_t &o, action_t a);

		// add a new (assumed unique) child node, returning the added child
		LZContextNode *addChild(LZContextNode *n);

		// do we have no children
		bool isLeaf(void) const;

		// computes Pr(obs_n | a_{1:n-1} obs_{1:n-1})
		double prob(void) const;

		// serialization routine
		template<class Archive>
		void serialize(Archive &ar, const unsigned int version);

	private:

		// only used by Boost Serialization
		LZContextNode(void) { }

		LZContextNode *m_parent;
		double m_prob;
		uint64_t m_count;
		double m_cost_to_go;
		lznode_list_t m_children;
		observation_t m_obs;
		action_t m_action;
};

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void LZContextNode::serialize(Archive &ar, const unsigned int version) {
	
	ar & m_parent;
	ar & m_prob;
	ar & m_count;
	ar & m_cost_to_go;
	ar & m_children;
	ar & m_obs;
	ar & m_action;
}



// Lempel-Ziv Context Tree
class LZContextTree {

	friend class boost::serialization::access;

	public:

		// create LZ-Context Tree from a file
		LZContextTree(const std::string &filename, bool use_binary_io);

		LZContextTree(unsigned int obs_bits, unsigned int rew_bits, 
			unsigned int n_actions, double discount);

		// the estimated probability of a percept occurring next
		// double prob(const context_t &context, const percept_t &p) const;

		// the estimated future discounted cost from a context
		double costToGo(const context_t &context, action_t a);

		// how many times has this context been seen before?
		uint64_t count(const context_t &context);

		// determine the estimated best action
		action_t bestAction(const context_t &context);

		// update the tree with new context information
		void update(const context_t &context);

		// extract the cost component from a percept
		double costFromPercept(const observation_t &obs) const;

		// number of distinct observations
		size_t nObservations(void) const;

		// serialization routine
		template<class Archive>
		void serialize(Archive &ar, const unsigned int version);

	private:

		// expected future cost if action a is executed from a certain context
		double cost(LZContextNode *n, action_t a) const;

		// attempts to match a context within the tree, NULL on no match
		LZContextNode *findNode(const context_t &context);

		// computes Pr(obs_n | a_{1:n-1} obs_{1:n-1}) from count information
		double probFromCounts(const LZContextNode &n) const;

		LZContextNode m_root;
		unsigned int m_obs_bits;
		unsigned int m_rew_bits;
		unsigned int m_n_actions;
		double m_discount;
};

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void LZContextTree::serialize(Archive &ar, const unsigned int version) {
	
	ar & m_root;
	ar & m_obs_bits;
	ar & m_rew_bits;
	ar & m_n_actions;
	ar & m_discount;
}


#endif // __ACTIVELZ_HPP__

