#include "activelz.hpp"
#include "log.hpp"

#include <iostream>
#include <limits>
#include <exception>
#include <stdexcept>
#include <string>
#include <fstream>

// boost includes
#include <boost/random.hpp>
#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>


namespace po = boost::program_options;

/// malformed environment response exception
class BadPerceptException : public std::exception { 
	const char *what(void) const throw() { return "invalid percept"; }
};

// program details
static const char program_name[] = "Active-LZ";
static const char version[]      = "alpha";
static const char authors[]      = "Joel Veness";


// maximum age of Active-LZ agent
static const unsigned int MaxAgentAge = 1000000000;


// stores the application configuration options and their description
po::variables_map options;
static po::options_description options_desc("Usage", 100);

static std::auto_ptr<LZContextTree> lzct;


/* load program options into a variable map */
static void initOptions(int argc, char *argv[], po::variables_map &vm) {

    options_desc.add_options()

        ("help", "")

	    ("reward-bits",
		 po::value<unsigned int>()->default_value(1),
		 "bits used to encode the reward signal")

	    ("observation-bits",
		 po::value<unsigned int>()->default_value(1),
         "bits used to encode the observation signal")

		 ("agent-actions",
		 po::value<unsigned int>()->default_value(4),
		 "the number of distinct actions the agent can do")

	    ("agent-load",
		 po::value<std::string>(),
		 "load a pre-existing agent from a file")

		("exploration",
		 po::value<double>()->default_value(0.05),
		 "probability of playing a random move")

	    ("discount",
		 po::value<double>()->default_value(0.99),
		 "discount factor")

		("explore-decay",
		 po::value<double>()->default_value(1.0),
		 "a value between [0.0,1.0] that defines the geometric decay of the exploration rate")

	    ("terminate-age",
		 po::value<age_t>()->default_value(MaxAgentAge),
		 "how many agent/environment cycles before the agent needs to close?")

		("binary-io",
		 "use native (faster but non-portable) binary file i/o")
    ;

    po::store(po::parse_command_line(argc, argv, options_desc), vm);
    po::notify(vm);
}


/* display the program usage options and quit. */
static void showHelp(void) {

	std::cout << std::endl;
	std::cout << program_name << " " << version << ", by " << authors << std::endl;
	std::cout << "-------------------------------" << std::endl << std::endl;
	std::cout << "The environment send percepts to the agent over stdin." << std::endl;
	std::cout << "The agent responds to the environment over stdout." << std::endl << std::endl;
	std::cout << "All communication channels are binary." << std::endl;
	std::cout << std::endl << options_desc << std::endl;
    exit(0);
}


/* checks for illegal combinations of configuration parameters,
   throwing an exception if an illegal combination is detected. */
static void processOptions(const po::variables_map &vm) {

	if (vm.count("help") > 0) showHelp();
    
	if (vm.count("observation-bits") > 0) {
		if (vm["observation-bits"].as<unsigned int>() < 1) {
			throw std::invalid_argument("the number of observation bits must be positive.");
		}
	}

	if (vm.count("reward-bits") > 0) {
		if (vm["reward-bits"].as<unsigned int>() < 1) {
			throw std::invalid_argument("the number of reward bits must be positive.");
		}
	}

	if (vm["reward-bits"].as<unsigned int>() + vm["observation-bits"].as<unsigned int>() > 30) {
		throw std::invalid_argument("the numbers of percept bits must be 30 or less");
	}


	if (vm.count("discount") > 0) {
		double gamma = vm["discount"].as<double>();
		if (gamma <= 0.0 || gamma >= 1.0) {
			throw std::invalid_argument("the discount factor must be within (0..1)");
		}
	}

	if (vm.count("terminate-age") > 0) {
		if (vm["terminate-age"].as<unsigned int>() < 1) {
			throw std::invalid_argument("the termination age must be positive.");
		}
	}
}


/* determine the number of bits needed to encode each action */
static unsigned int num_action_bits(void) {

	unsigned int actions_c = options["agent-actions"].as<unsigned int>();
	unsigned int actions_bits_c;

	for (unsigned int i = 1, c = 1; i < actions_c; i *= 2, c++) {
		actions_bits_c = c;
	}

	return actions_bits_c;
}


/* save the state of an agent to disk, false on failure */
bool saveAgent(const LZContextTree &lzct, const std::string &filename) {

	if (options.count("binary-io") > 0) {

		std::ofstream ofs(filename.c_str(), std::ios_base::binary);
		if (!ofs) return false;
 
		boost::archive::binary_oarchive oa(ofs);
        oa << lzct;

	} else {
		
		std::ofstream ofs(filename.c_str());
		if (!ofs) return false;
    
		boost::archive::text_oarchive oa(ofs);
        oa << lzct;
    }

	LINFO_ << "successfully saved agent to: " << filename;
	return true;
}


/* handle user commands (prefixed by colons), returning
   false if input is not a user command. */
static bool dispatchCommand(const std::string &cmd) {

	typedef boost::tokenizer<boost::char_separator<char> > Tok;
	boost::char_separator<char> sep(" ");
	Tok tok(cmd, sep);

	if (cmd.size() == 0 || cmd[0] != ':') return false;

	if (cmd == ":quit") {

		std::exit(0);

	} else if (cmd.substr(0, 5) == ":save") {

		Tok::iterator it = tok.begin();
		std::string filename = "";
		if (++it != tok.end()) filename = *it;
		
		assert(lzct.get() != NULL);
		bool rval = saveAgent(*lzct, filename);
		if (!rval) {
			LINFO_ << "failed to save agent at " << filename;
		}

	} else {

		LINFO_ << "unrecognised user command, ignoring...";
		return false;
	}

	return true;
}


/* get an action uniformly randomly */
static action_t getUniformRandomAction(void) {

	static unsigned int n_actions = options["agent-actions"].as<unsigned int>();

	action_t a = action_t(urng() * double(n_actions));
	return a;
}


/* agent <-> environment communication loop */
static void mainLoop(void) {

	unsigned int obs_bits  = options["observation-bits"].as<unsigned int>();
	unsigned int rew_bits  = options["reward-bits"].as<unsigned int>();
	double discount        = options["discount"].as<double>();
	double exploration     = options["exploration"].as<double>();
	double exp_decay       = options["explore-decay"].as<double>();
	unsigned int n_actions = options["agent-actions"].as<unsigned int>();
	unsigned int act_bits  = num_action_bits();
	unsigned int term_age  = options["terminate-age"].as<unsigned int>();
	
	unsigned int percept_bits = obs_bits + rew_bits;
	double total_reward = 0.0;
	std::string buf;
	action_t action = NullAction;
	
	unsigned int t   = 1;  // timestep
	unsigned int c   = 1;  // current phrase index
	unsigned int tau = 1;  // start time of current phrase

	// create Active-LZ context tree
	if (options.count("agent-load") > 0) {
		bool use_binary_io = options.count("binary-io") > 0;
		lzct.reset(new LZContextTree(options["agent-load"].as<std::string>(), use_binary_io));
	} else {
		lzct.reset(new LZContextTree(obs_bits, rew_bits, n_actions, discount));
	}

	context_t context;

	do {
		if (!std::getline(std::cin, buf)) break;

		// process any user directives
		if (dispatchCommand(buf)) continue;

		// grab percept
		boost::dynamic_bitset<> p = boost::dynamic_bitset<>(buf);
		if (p.size() != percept_bits) throw BadPerceptException();

		// add <obs_n, action_{n-1} to context
		context.push_back(obs_action_t(action, p));

		// split into observation and reward components
		double reward = -lzct->costFromPercept(p);
		total_reward += reward;

		LINFO_ << "Received reward of " << reward << " on timestep " << t;
		double avg_reward = total_reward / double(t);
		LINFO_ << "Average session reward: " << avg_reward;

		uint64_t count = lzct->count(context);
		if (count == 0) {
			LINFO_ << "Currently in unexplored context:";
		} else {
			LINFO_ << "Currently in context visited " << count << " times:";
		}

		for (action_t a=0; a < n_actions; a++) {
			LINFO_ << "  action: " << a << "   Cost-to-go: " << lzct->costToGo(context, a);
		}

		// if we currently are in a context we have seen before
		if (lzct->count(context) > 0) {

			if (urng() < exploration) {
				action = getUniformRandomAction();
				LINFO_ << "Exploring action: " << action;
			} else {
				action = lzct->bestAction(context);
				LINFO_ << "Greedily choosing action: " << action;
			}

		} else {

			action = getUniformRandomAction();
			LINFO_ << "Exploring action: " << action;

			lzct->update(context);
			c++;
			tau = t + 1;
			context.clear();
		}

		// send response to the environment
		boost::dynamic_bitset<> response(act_bits, action);
		response.resize(act_bits);
		std::cout << response << std::endl;

		exploration *= exp_decay;
		t++;

	} while (t <= term_age);

	// display session summary
	std::ostringstream oss;
	oss << "Session summary: " << total_reward 
		<< " reward from " << (t-1) << " cycles.";
	LINFO_ << oss.str();
	std::cout << oss.str();
}


/* application entry point */
int main(int argc, char *argv[]) {

	try {

		initLogs();

		// load program configuration
        initOptions(argc, argv, options); 

        // check and apply program configuration
        processOptions(options);

		// enter agent <-> environment communication loop
		mainLoop();
		
	} catch (std::exception &e) {
    
		LERR_ << "error: " << e.what() << std::endl;
        return 1;

    } catch (...) {

		LERR_ << "error: unknown exception" << std::endl;
        return 1;
    }

	LINFO_ << "agent terminated";

	return 0;
}

