#include "kstest.hpp"

#include <cmath>
#include <algorithm>
#include <cassert>
#include <limits>

#include <boost/math/special_functions/fpclassify.hpp>


/* Kolmogorov-Smirnov probability function */
static double probks(double alam) {

	static const double Eps1 = 0.001;
	static const double Eps2 = 1.0e-8;

	double fac = 2.0, sum = 0.0, termbf = 0.0;
	double a2 = -2.0 * alam * alam;

	for (int j=1; j <= 100; j++) {
		
		double term = fac * std::exp(a2 * j * j);
		sum += term;

		if (std::fabs(term) <= Eps1 * termbf || 
			std::fabs(term) <= Eps2 * sum) return sum;

		fac = -fac;
		termbf = std::fabs(term);
	}

	return 1.0;
}


/* Given an array data1[1..n1], and an array data2[1..n2],
   this routine returns the K-S statistic d, and the significance
   level prob for the null hypothesis that the data sets are drawn
   from the same distribution. Small values of prob show that the
   cumulative distribution function of data1 is significantly different
   from that of data2. Assumes that data1 and data2 are sorted. */
static void kstwo_sorted_data(
	const double data1[], size_t n1, 
	const double data2[], size_t n2, 
	double &d, 
	double &prob
) {

	size_t j1 = 0, j2 = 0;
	double fn1 = 0.0, fn2 = 0.0;
	double en1 = n1, en2 = n2;
	d = 0.0;

	while (j1 < n1 && j2 < n2) {

		while (j1 < n1-1 && data1[j1] == data1[j1+1]) j1++;
		while (j2 < n2-1 && data2[j2] == data2[j2+1]) j2++;

		double d1 = data1[j1], d2 = data2[j2];
		assert(!boost::math::isnan(d1) && !boost::math::isnan(d2));

		if (d1 <= d2) fn1 = double(++j1) / en1;
		if (d2 <= d1) fn2 = double(++j2) / en2;

		double dt = std::fabs(fn2-fn1);
		if (dt > d) d = dt;
	}
	double en = std::sqrt(en1 * en2 / (en1 + en2));
	prob = probks((en + 0.12 + 0.11 / en) * d);
}


/* c++ wrapper function for Kolmogorov-Smirnov test */
void kolmogorov_smirnov_test(
	const std::vector<double> &data1, 
	const std::vector<double> &data2, 
	double &d, 
	double &prob
) {

	if (data1.size() == 0 || data2.size() == 0) {
		d = std::numeric_limits<double>::signaling_NaN();
		prob = 1.0;
		return;
	}

	assert(data1.size() > 0);
	assert(data2.size() > 0);

	std::vector<double> d1(data1);
	std::vector<double> d2(data2);

	std::sort(d1.begin(), d1.end());
	std::sort(d2.begin(), d2.end());
	
	kstwo_sorted_data(&d1[0], d1.size(), &d2[0], d2.size(), d, prob);
}

