#include "utree.hpp"
#include "log.hpp"
#include "kstest.hpp"

#include <iostream>
#include <limits>
#include <exception>
#include <stdexcept>
#include <map>
#include <sstream>
#include <fstream>

#include <boost/random.hpp>
#include <boost/math/special_functions/fpclassify.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>

#undef max


// node types
static const int NodeSplit  = 0;
static const int NodeLeaf   = 1;
static const int NodeFringe = 2;

static const int ActionDimension = -1;

// a node must have at least this many instances to be a split candidate
static const int MinSplitInstances = 10;

// significance values below this threshold are considered to be
// indicative of different reward distributions
static const double SignificanceLevel = 0.0001;


// uniform [0..1) random number generation
static randsrc_t rng;
static randgen_t urng(rng);


/* send an instance to a stream */
std::ostream &operator<<(std::ostream &o, const instance_t &inst) {

	static const std::string Sep = "/";
	std::ostringstream oss;
	
	oss << "a_t-1=" << inst.action << Sep
		<< "o_t=" << inst.obs << Sep
		<< "r_t=" << inst.reward << Sep
	    << "t=" << inst.timestep;
	return o << "Instance[" << oss.str() << "]";
}


/* send a distinction to a stream */
std::ostream &operator<<(std::ostream &o, const distinction_t &d) {
	
	std::ostringstream oss;
	
	oss << d.back_idx << ":";
	if (d.dimension == ActionDimension) {
		oss << "Action";
	} else {
		oss << "Obs/" << d.dimension;
	}
		
	return o << "Distinction[" << oss.str() << "]";
}


/* send an action model to a stream */
std::ostream &operator<<(std::ostream &o, const action_model_t &am) {

	static const std::string Sep = "/";

	o << "ActionModel[c=" << am.count << Sep
	  << "r=" << am.reward << Sep
	  << "q=" << am.q_value << Sep;

	trans_cnt_t::const_iterator it = am.trans_cnt.begin();
	for (; it != am.trans_cnt.end(); ++it) {
		double prob = double(it->second) / double(am.count);
		o << "to(" << it->first << ")=" << it->second << "/" << am.count << ",";
	}
	o << "]";

	return o;
}


/* send a UNode to a stream */
std::ostream &operator<<(std::ostream &o, const UNode &un) {

	static const std::string Sep = "/";

	std::ostringstream oss, parent;
	if (un.parent() != NULL) parent << un.parent()->id();

	oss << un.id() << ":"
		<< "p=" << parent.str() << Sep
		//<< "r=" << un.expectedReward(un.bestAction()) << Sep
		//<< "a*=" << un.bestAction() << Sep
		//<< "u=" << un.utility() << Sep
		<< "# children=" << un.nChildren() << Sep
		<< "# instances=" << un.nInstances();

	model_t::const_iterator it = un.m_model.begin();
	for (; it != un.m_model.end(); ++it) {
		oss << Sep << it->first << ":" << it->second;
	}

	return o << "UNode[" << oss.str() << "]"; 
}


/* sends a U-Tree to a stream */
std::ostream &operator<<(std::ostream &o, const UTree &utree) {

	utree.m_root.childrenToStream(o);
	return o;
}



/* create a new UNode */
UNode::UNode(int id, int type, UNode *parent) :
	m_id(id), m_type(type), m_parent(parent)
{
}


/* compute the expected immediate reward by taking some action */
double UNode::expectedReward(action_t action) const {

	model_t::const_iterator it = m_model.find(action);
	if (it == m_model.end()) return 0.0; // HACK: check what to do?

	double total_rew = double(it->second.reward);
	double cnt       = double(it->second.count);

	return total_rew / cnt;
}


/* number of children nodes */
size_t UNode::nChildren(void) const {
	
	return m_children.size();
}


/* determine the parent of a node */
const UNode *UNode::parent(void) const {
	
	return m_parent;
}


/* number of instances within a node */
size_t UNode::nInstances(void) const {
	
	return m_instances.size();
}

/* compute the probabililty of transitioning to particular
   state if a certain action is taken */
double UNode::transProb(unode_id_t id, action_t action) const {

	model_t::const_iterator it = m_model.find(action);
	if (it == m_model.end()) return 0.0;
	if (it->second.count == 0) return 0.0;

	trans_cnt_t::const_iterator i = it->second.trans_cnt.find(id);
	if (i == it->second.trans_cnt.end()) return 0.0;

	assert(it->second.count >= i->second);
	assert(it->second.count > 0);

	return double(i->second) / double(it->second.count);
}


/* compute the expected utility of a leaf */
double UNode::utility(void) const {

	if (m_model.empty()) return 0.0;

	double utility = -std::numeric_limits<double>::infinity();

	model_t::const_iterator it = m_model.begin();
	for (; it != m_model.end(); ++it) {
		int action = it->first;
		utility = std::max(utility, it->second.q_value);
	}

	return utility;
}


/* adds an instance to the current leaf node */
void UNode::insert(const instance_t &instance) {

	assert(type() == NodeLeaf);
	m_instances.push_back(instance);
}


/* determine the relevant child of a split node by
   applying a utile distinction */
size_t UNode::applyDistinction(const history_t &history, int idx) const {

	assert(m_type == NodeSplit);
	assert(int(history.size()) > m_distinction.back_idx);
	assert(m_distinction.back_idx >= 0);
	assert(idx < int(history.size()));

	if (idx == -1 || idx == 0) return 0;
	assert(idx > 0);

	// handle degenerate case: 
	// when back_idx goes too far back in time for the current instance,
	// which will only happen for a small number of instance, just pick the
	// first child
	if (m_distinction.back_idx > idx) return 0;

	const instance_t &inst = history[idx - m_distinction.back_idx];

	if (m_distinction.dimension == ActionDimension) {		
		return inst.action;
	}

	// otherwise test the relevant observation component
	assert(m_distinction.dimension < int(inst.obs.size())); 
	return inst.obs[m_distinction.dimension];
}


/* get the leaf matching the current state */
UNode *UNode::getLeaf(const history_t &history) const {

	assert(m_type != NodeFringe);

	if (m_type == NodeLeaf) return const_cast<UNode *>(this);

	assert(m_type == NodeSplit);
	assert(m_children.size() > 0);

	size_t child_idx = applyDistinction(history, history.size() - 1);
	return m_children[child_idx].getLeaf(history); 
}


/* recursively perform a sweep of value iteration on all
   the leaf nodes contained in the current subtree */
size_t UNode::sweep(double discount) {

	if (m_type == NodeLeaf) {
		
		// perform a single step of value iteration on each q-value
		model_t::iterator it = m_model.begin();
		for (; it != m_model.end(); ++it) {
			
			action_t a = it->first;

			double exp_utility = 0.0;
			trans_cnt_t::iterator i = it->second.trans_cnt.begin();
			for (; i != it->second.trans_cnt.end(); ++i) {
				exp_utility += transProb(i->first, a) * utility();			
			}

			it->second.q_value = expectedReward(a) + discount * exp_utility;
		}
		
		return 1;
	}

	assert(m_type == NodeSplit);

	size_t total = 0;
	for (size_t i=0; i < m_children.size(); i++) {
		total += m_children[i].sweep(discount);
	}

	return total;
}


/* determine the best action */
action_t UNode::bestAction(void) const {

	//assert(m_type == NodeLeaf);

	std::vector<action_t> best_actions;
	double best_q = -std::numeric_limits<double>::infinity();

	model_t::const_iterator it = m_model.begin();
	for (; it != m_model.end(); ++it) {
		if (it->second.q_value > best_q) best_actions.clear();
		if (it->second.q_value >= best_q) {
			best_actions.push_back(it->first);
			best_q = it->second.q_value;
		}	
	}

	assert(best_actions.size() > 0);
	int rand_idx = int(urng() * double(best_actions.size()));
	return best_actions[rand_idx];
}


/* determine the id of a node */
unode_id_t UNode::id(void) const {

	return m_id;
}


/* update the transition and reward models */
void UNode::updateModel(action_t a, unode_id_t id, double r) {

	action_model_t &am = m_model[a];
	am.trans_cnt[id]++;
	am.count++;
	am.reward += r;

	assert(am.trans_cnt[id] <= am.count);
}


/* recursively send information about children nodes to stream */
void UNode::childrenToStream(std::ostream &o) const {

	if (m_type == NodeLeaf) o << *this << std::endl;

	for (size_t i=0; i < m_children.size(); i++) {
		m_children[i].childrenToStream(o);
	}
}


/* the type of UNode */
int UNode::type(void) const {

	return m_type;
}



/* create a U-Tree data structure */
UTree::UTree(
	unsigned int obs_bits, 
	unsigned int n_actions, 
	unsigned int max_back,
	double discount
) :
	m_root(0, NodeLeaf, NULL),
	m_node_id_c(1),
	m_obs_bits(obs_bits),
	m_n_actions(n_actions),
	m_max_back_depth(max_back),
	m_discount(discount)	
{
	m_nodes[m_root.id()] = &m_root;
}


/* create a U-Tree data structure from a file */
UTree::UTree(const std::string &filename, bool use_binary_io) :
	m_root(0, NodeLeaf, NULL)
{
	if (use_binary_io) {

		// load the agent from file
		std::ifstream ifs(filename.c_str(), std::ios_base::binary);
		if (!ifs) throw std::invalid_argument("bad UTree agent file.");
		LINFO_ << "loading (binary-mode) UTree agent from: " << filename << std::endl;

		boost::archive::binary_iarchive ia(ifs);
        ia >> *this;

	} else { 
		
		// load the agent from file
		std::ifstream ifs(filename.c_str());
		if (!ifs) throw std::invalid_argument("bad UTree agent file.");
		LINFO_ << "loading (text-mode) UTree agent from: " << filename << std::endl;

		boost::archive::text_iarchive ia(ifs);
        ia >> *this;
    }
}


/* perform a single sweep of value iteration, returning
   the number of nodes updated */
size_t UTree::sweep(void) {

	return m_root.sweep(m_discount);
}


/* processes a new instance */
void UTree::update(const instance_t &instance) {
	
	UNode *old_state = getLeaf();

	// add instance to relevant leaf
	old_state->insert(instance);

	// append to history
	m_history.push_back(instance);

	// update transition and reward models
	UNode *new_state = getLeaf();
	old_state->updateModel(instance.action, new_state->id(), instance.reward);
}


/* return the best action */
action_t UTree::bestAction(void) const {

	const UNode *l = getLeaf();

	std::vector<action_t> unexplored;
	for (action_t a = 0; a < nActions(); a++) {
		model_t::const_iterator it = l->m_model.find(a);
		if (it == l->m_model.end() || it->second.count == 0) {
			unexplored.push_back(a);
		}
	}

	if (unexplored.size() == nActions()) {
		double r = urng() * double(unexplored.size());
		LINFO_ << "forcing exploration";
		return unexplored[size_t(r)];
	}

	return l->bestAction();
}


/* determine the current node corresponding the the agent's state*/
UNode *UTree::getLeaf(void) const {

	return m_root.getLeaf(m_history);
}


/* get the current model statistics for a particular action */
const action_model_t *UTree::actionModel(action_t a) const {

	const UNode *l = getLeaf();

	if (l->m_model.find(a) == l->m_model.end()) return NULL;
	return &l->m_model.find(a)->second;
}


/* expand the fringe, keeping the expansion if it helps */
unsigned int UTree::trySplits(unsigned int fringe) {

	return trySplitRecursive(m_root, fringe);
}


/* the size of the action space */
unsigned int UTree::nActions(void) const {
	
	return m_n_actions;
}


/* how many observation bits in each percept */
unsigned int UTree::nObsBits(void) const {
	
	return m_obs_bits;
}


/* maximum depth back that a utile distinction can be made */
unsigned int UTree::maxBackDepth(void) const {

	return m_max_back_depth;
}


/* generate a new unique id */
unode_id_t UTree::genId(void) {

	return ++m_node_id_c;
}


/* transform a leaf node into a split node */
void UTree::split(UNode &un, const distinction_t &d) {

	assert(un.type() == NodeLeaf);
	assert(d.back_idx >= 0);
	assert(un.m_children.size() == 0);

	un.m_type = NodeSplit;
	un.m_distinction = d;

	// add children
	if (d.dimension == ActionDimension) {
		for (unsigned int i=0; i < nActions(); i++) {
			unode_id_t id = genId();
			m_nodes[id] = new UNode(id, NodeLeaf, &un);
			un.m_children.push_back(m_nodes[id]);
		}
	} else {
		unode_id_t id1 = genId(), id2 = genId();
		m_nodes[id1] = new UNode(id1, NodeLeaf, &un);
		m_nodes[id2] = new UNode(id2, NodeLeaf, &un);
		un.m_children.push_back(m_nodes[id1]);
		un.m_children.push_back(m_nodes[id2]);
	}

	// add the instances to the children
	for (size_t i=0; i < un.m_instances.size(); i++) {
		UNode *n = getInstanceLeaf(un.m_instances[i]);
		assert(n->parent()->id() == un.id());
		n->insert(un.m_instances[i]);
	}

	// re-build the tree model
	node_list_t::iterator it = m_nodes.begin();
	for (; it != m_nodes.end(); ++it) {
		if (it->second->type() == NodeLeaf) {
			modelFromInstances(*it->second);
		}
	}

	// create q-values of the children
	for (size_t i=0; i < un.m_children.size(); i++) {
		un.m_children[i].sweep(m_discount);
	}
}


/* transform a (recently) split node into a leaf node, inverse of split */
void UTree::unsplit(UNode &un) {

	assert(un.m_type == NodeSplit);
	assert(un.m_children.size() > 0);

	un.m_type = NodeLeaf;
	un.m_distinction = distinction_t();

	for (size_t i=0; i < un.m_children.size(); i++) {
		m_nodes.erase(un.m_children[i].id());
	}

	un.m_children.clear();

	// re-build the tree model
	node_list_t::iterator it = m_nodes.begin();
	for (; it != m_nodes.end(); ++it) {
		if (it->second->type() == NodeLeaf) modelFromInstances(*it->second);
	}
}


/* attempts to split nodes according to a utile distinction test,
   returning the number of nodes split beneath the current subtree */
unsigned int UTree::trySplitRecursive(UNode &un, unsigned int fringe) {

	// TODO: handle fringe parameter

	// skip splitting if node has not enough instances
	if (un.nInstances() < MinSplitInstances) return 0;

	if (un.m_type == NodeLeaf) {
		
		// generate a set of candidate distinctions
		distinction_list_t candidates;
		genSplitCandidates(un, candidates);

		distinction_t *d = getUtileDistinction(un, candidates);
		if (d) {
			LINFO_ << "Worthwhile " << *d << " found at node " << un.id();
			split(un, *d);
			// now try to further split the node
			return 1 + trySplitRecursive(un, fringe);
		}
		return 0;
	}

	// recursively split children
	unsigned int total = 0;
	for (size_t i=0; i < un.m_children.size(); i++) {
		total += trySplitRecursive(un.m_children[i], fringe);
	}

	return total;
}


/* generate all possibile utile distinctions */
void UTree::genSplitCandidates(const UNode &un, distinction_list_t &candidates) const {

	candidates.clear();

	distinction_list_t occurred;
	ancestorDistinctions(un, occurred);

	int n = nObsBits();

	// order from most recent to latest
	for (unsigned int i=0; i <= maxBackDepth(); i++) {
		for (int j=-1; j < n; j++) {
			distinction_t d = distinction_t(i, j);
			if (std::find(occurred.begin(), occurred.end(), d) == occurred.end()) {
				candidates.push_back(d);
			}
		}
	}
}


/* get the list of ancestor distinctions */
void UTree::ancestorDistinctions(const UNode &un, distinction_list_t &occurred) const {

	occurred.clear();
	const UNode *p = un.parent();

	while (p != NULL) {
		assert(p->m_type == NodeSplit);
		occurred.push_back(p->m_distinction);
		p = p->parent();
	}
}


/* the expected future discounted reward of all of a nodes instances */
void UTree::getEDFRList(UNode &un, efdr_list_t &efdr_list) {
	
	efdr_list.clear();

	for (size_t i=0; i < un.m_instances.size(); i++) {
		double efdr = expectedFutureReward(un.m_instances[i]);
		assert(!boost::math::isnan(efdr));
		efdr_list.push_back(efdr);
	}
}


/* attempt to find a worthwhile utile distinction from a current node */
distinction_t *UTree::getUtileDistinction(UNode &un, distinction_list_t &dl) {

	// determine the set of root expected future discounted rewards
	efdr_list_t root_utilities;
	getEDFRList(un, root_utilities);

	// TODO: consider randomizing the list of distinctions, to remove bias?
	//       or at least make the ordering a parameter to the program
	
	// check to see whether any utile distinctions are helpful
	for (size_t i=0; i < dl.size(); i++) {

		split(un, dl[i]);

		// compute expected future rewards for each instance
		std::vector<std::vector<double> > leaf_utilities(un.nChildren());
		for (size_t j=0; j < un.nChildren(); j++) {
			getEDFRList(un.m_children[j], leaf_utilities[j]);
		}

		unsplit(un);
		
		// do any of the leaves show significantly different expected
		// future reward distributions to the root?
		for (size_t j=0; j < leaf_utilities.size(); j++) {
			double d, prob;
			kolmogorov_smirnov_test(root_utilities, leaf_utilities[j], d, prob);
			//LINFO_ << "Split condifence: " << (1.0 - prob) << std::endl;
			if (prob < SignificanceLevel) return &dl[i];
		}
	}

	return NULL;
}


/* expected future discounted reward associated with an instance */
double UTree::expectedFutureReward(const instance_t &inst) {

	double exp_future_rew = inst.reward, total = 0.0;

	UNode *un = getInstanceLeaf(inst);

	// find the relevant action statistics
	if (un->m_model.size() == 0) return 0.0;
	action_model_t &am = un->m_model[un->bestAction()];
		
	trans_cnt_t::const_iterator it = am.trans_cnt.begin();
	for (; it != am.trans_cnt.end(); ++it) {
		double prob = un->transProb(it->first, inst.action);
		double util = m_nodes[it->first]->utility();
		assert(!boost::math::isnan(prob));
		assert(!boost::math::isnan(util));
		total += prob * util;
		assert(!boost::math::isnan(total));
	}
	exp_future_rew += m_discount * total;

	return exp_future_rew;
}


/* get the leaf node associated with an instance */
UNode *UTree::getInstanceLeaf(const instance_t &inst) {

	UNode *n = &m_root;

	// apply distinction to the the history proceeding the instance
	int idx = (inst.timestep - 1) - 1;

	while (n->type() != NodeLeaf) {
		assert(n->type() == NodeSplit);
		size_t c = n->applyDistinction(m_history, idx);
		n = &n->m_children[c];
	}

	return n;
}


/* access the root U-Node */
const UNode &UTree::root(void) const {
	
	return m_root;
}


/* the id of the current state */
unode_id_t UTree::state(void) const {

	return getLeaf()->id();
}


/* compute the transition probability for a node from scratch */
double UTree::transProbFromInstances(const UNode &un, unode_id_t id, action_t action) {

	double c = 0.0, total = 0.0;

	for (size_t i=0; i < un.nInstances(); i++) {

		const instance_t &inst = un.m_instances[i];
		UNode *from = getInstanceLeaf(inst);
		const instance_t *ninst = nextInstance(inst);
		UNode *to = ninst == NULL ? getLeaf() : getInstanceLeaf(*ninst);
	
		if (inst.action == action) {
			if (to->id() == id) c += 1.0;
			total += 1.0; 
		}
	}

	return (total != 0.0) ? c / total : 0.0;
}


/* compute the instanteous reward for a node from scratch */
double UTree::rewardFromInstances(const UNode &un, action_t action) {

	double rew_total = 0.0, total = 0.0;

	for (size_t i=0; i < un.nInstances(); i++) {

		if (un.m_instances[i].action == action) {
			rew_total += un.m_instances[i].reward;
			total += 1.0;
		}
	}

	return total != 0.0 ? rew_total / total : 0.0;
}


/* compute the model from the instances */
void UTree::modelFromInstances(UNode &un) {

	double c = 0.0, total = 0.0;

	// clear transition and reward statistics, but 
	// importantly, preserve the q-values
	model_t::iterator it = un.m_model.begin();
	for (; it != un.m_model.end(); ++it) {
		it->second.count = 0;
		it->second.reward = 0;
		it->second.trans_cnt.clear();
	}

	for (size_t i=0; i < un.nInstances(); i++) {

		const instance_t &inst = un.m_instances[i];

		// what state were we in when we received this instance?
		UNode *from = getInstanceLeaf(inst);
		const instance_t *ninst = nextInstance(inst);
		UNode *to = ninst == NULL ? getLeaf() : getInstanceLeaf(*ninst);

		from->updateModel(inst.action, to->id(), inst.reward);
	}
}


/* the instance that was received after the current one */
const instance_t *UTree::nextInstance(const instance_t &inst) const {
	
	return inst.timestep < m_history.size() ? &m_history[inst.timestep] : NULL;
}


/* checks the integrity of the u-tree, false on failure */
bool UTree::integrityTest(void) {

	return integrityTestWorker(m_root);
}


/* checks the integrity of the u-tree, false on failure */
bool UTree::integrityTestWorker(UNode &n) {

	if (n.type() == NodeLeaf) {
		
		for (action_t a=0; a < nActions(); a++) {
			node_list_t::const_iterator it = m_nodes.begin();
			for (; it != m_nodes.end(); ++it) {
				double tp = n.transProb(it->first, a);
				double tp_slow = transProbFromInstances(n, it->first, a);
				if (tp != tp_slow) return false;
			}
			double rew = n.expectedReward(a);
			double rew_slow = rewardFromInstances(n, a);
			if (rew != rew_slow) return false;
		}
		return true;
	}

	for (size_t i=0; i < n.m_children.size(); i++) {
		if (!integrityTestWorker(n.m_children[i])) return false;
	}
	return true;
}


